package net.minecraftforge.client.model;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import net.minecraft.block.Block;
import net.minecraft.block.state.IBlockState;
import net.minecraft.client.Minecraft;
import net.minecraft.client.renderer.BlockModelShapes;
import net.minecraft.client.renderer.ItemMeshDefinition;
import net.minecraft.client.renderer.ItemModelMesher;
import net.minecraft.client.renderer.block.model.BlockPart;
import net.minecraft.client.renderer.block.model.BlockPartFace;
import net.minecraft.client.renderer.block.model.ItemCameraTransforms;
import net.minecraft.client.renderer.block.model.ItemModelGenerator;
import net.minecraft.client.renderer.block.model.ModelBlock;
import net.minecraft.client.renderer.block.model.ModelBlockDefinition;
import net.minecraft.client.renderer.block.model.ModelBlockDefinition.MissingVariantException;
import net.minecraft.client.renderer.block.model.ModelBlockDefinition.Variant;
import net.minecraft.client.renderer.block.model.ModelBlockDefinition.Variants;
import net.minecraft.client.renderer.block.statemap.IStateMapper;
import net.minecraft.client.renderer.texture.IIconCreator;
import net.minecraft.client.renderer.texture.TextureAtlasSprite;
import net.minecraft.client.renderer.texture.TextureMap;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.client.renderer.vertex.VertexFormat;
import net.minecraft.client.resources.IResourceManager;
import net.minecraft.client.resources.model.BuiltInModel;
import net.minecraft.client.resources.model.IBakedModel;
import net.minecraft.client.resources.model.ModelBakery;
import net.minecraft.client.resources.model.ModelResourceLocation;
import net.minecraft.client.resources.model.ModelRotation;
import net.minecraft.client.resources.model.SimpleBakedModel;
import net.minecraft.client.resources.model.WeightedBakedModel;
import net.minecraft.item.Item;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.IRegistry;
import net.minecraft.util.ResourceLocation;
import net.minecraftforge.fml.common.FMLLog;
import net.minecraftforge.fml.common.registry.GameData;
import net.minecraftforge.fml.common.registry.RegistryDelegate;

import org.apache.commons.lang3.tuple.Pair;

import com.google.common.base.Function;
import com.google.common.base.Throwables;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Lists;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;

import net.minecraft.client.renderer.block.model.ModelBlockDefinition.MissingVariantException;
import net.minecraft.client.renderer.block.model.ModelBlockDefinition.Variant;
import net.minecraft.client.renderer.block.model.ModelBlockDefinition.Variants;
@SuppressWarnings("deprecation")
public class ModelLoader extends ModelBakery
{
    private final Map<ModelResourceLocation, IModel> stateModels = new HashMap<ModelResourceLocation, IModel>();
    private final Set<ResourceLocation> textures = new HashSet<ResourceLocation>();
    private final Set<ResourceLocation> loadingModels = new HashSet<ResourceLocation>();
    private final Set<ModelResourceLocation> missingVariants = Sets.newHashSet();
    private IModel missingModel = null;
    private IModel itemModel = new ItemLayerModel(field_177606_o);

    private boolean isLoading = false;
    public boolean isLoading()
    {
        return isLoading;
    }

    public ModelLoader(IResourceManager manager, TextureMap map, BlockModelShapes shapes)
    {
        super(manager, map, shapes);
        VanillaLoader.instance.setLoader(this);
        ModelLoaderRegistry.clearModelCache();
    }

    @Override
    public IRegistry<ModelResourceLocation, IBakedModel> func_177570_a()
    {
        isLoading = true;
        loadBlocks();
        loadItems();
        try
        {
            missingModel = getModel(new ResourceLocation(field_177604_a.func_110624_b(), field_177604_a.func_110623_a()));
        }
        catch (IOException e)
        {
            // If this ever happens things are bad. Should never NOT be able to load the missing model.
            Throwables.propagate(e);
        }
        stateModels.put(field_177604_a, missingModel);
        textures.remove(TextureMap.field_174945_f);
        textures.addAll(field_177602_b);
        field_177609_j.func_174943_a(field_177598_f, new IIconCreator()
        {
            public void func_177059_a(TextureMap map)
            {
                for(ResourceLocation t : textures)
                {
                    map.func_174942_a(t);
                }
            }
        });
        Function<ResourceLocation, TextureAtlasSprite> textureGetter = new Function<ResourceLocation, TextureAtlasSprite>()
        {
            public TextureAtlasSprite apply(ResourceLocation location)
            {
                return Minecraft.func_71410_x().func_147117_R().func_110572_b(location.toString());
            }
        };
        IFlexibleBakedModel missingBaked = missingModel.bake(missingModel.getDefaultState(), DefaultVertexFormats.field_176599_b, textureGetter);
        for (Entry<ModelResourceLocation, IModel> e : stateModels.entrySet())
        {
            if(e.getValue() == getMissingModel())
            {
                field_177605_n.func_82595_a(e.getKey(), missingBaked);
            }
            else
            {
                field_177605_n.func_82595_a(e.getKey(), e.getValue().bake(e.getValue().getDefaultState(), DefaultVertexFormats.field_176599_b, textureGetter));
            }
        }
        return field_177605_n;
    }

    private void loadBlocks()
    {
        Map<IBlockState, ModelResourceLocation> stateMap = field_177610_k.func_178120_a().func_178446_a();
        Collection<ModelResourceLocation> variants = Lists.newArrayList(stateMap.values());
        variants.add(new ModelResourceLocation("minecraft:item_frame", "normal")); //Vanilla special cases item_frames so must we
        variants.add(new ModelResourceLocation("minecraft:item_frame", "map"));
        func_177591_a(variants);
    }

    @Override
    protected void func_177569_a(ModelBlockDefinition definition, ModelResourceLocation location)
    {
        Variants variants = null;
        try
        {
            variants = definition.func_178330_b(location.func_177518_c());
        }
        catch(MissingVariantException e)
        {
            missingVariants.add(location);
        }
        if (variants != null && !variants.func_178420_b().isEmpty())
        {
            try
            {
                stateModels.put(location, new WeightedRandomModel(location, variants));
            }
            catch(Throwable e)
            {
                throw new RuntimeException(e);
            }
        }
    }

    private void loadItems()
    {
        func_177592_e();
        for(Item item : GameData.getItemRegistry().typeSafeIterable())
        {
            for(String s : (List<String>)func_177596_a(item))
            {
                ResourceLocation file = func_177583_a(s);
                ModelResourceLocation memory = new ModelResourceLocation(s, "inventory");
                IModel model = null;
                try
                {
                    model = getModel(file);
                }
                catch (IOException e)
                {
                    // Handled by our finally block.
                }
                finally
                {
                    if (model == null || model == getMissingModel())
                    {
                        FMLLog.fine("Item json isn't found for '" + memory + "', trying to load the variant from the blockstate json");
                        func_177569_a(func_177586_a(memory), memory);
                    }
                    else stateModels.put(memory, model);
                }
            }
        }
    }

    public IModel getModel(ResourceLocation location) throws IOException
    {
        if(!ModelLoaderRegistry.loaded(location)) loadAnyModel(location);
        return ModelLoaderRegistry.getModel(location);
    }

    @Override
    protected ResourceLocation func_177580_d(ResourceLocation model)
    {
        return new ResourceLocation(model.func_110624_b(), model.func_110623_a() + ".json");
    }

    private void loadAnyModel(ResourceLocation location) throws IOException
    {
        if(loadingModels.contains(location))
        {
            throw new IllegalStateException("circular model dependencies involving model " + location);
        }
        loadingModels.add(location);
        try
        {
            IModel model = ModelLoaderRegistry.getModel(location);
            for (ResourceLocation dep : model.getDependencies())
            {
                getModel(dep);
               
            }
            textures.addAll(model.getTextures());
        }
        finally
        {
            loadingModels.remove(location);
        }
    }

    private class VanillaModelWrapper implements IRetexturableModel
    {
        private final ResourceLocation location;
        private final ModelBlock model;

        public VanillaModelWrapper(ResourceLocation location, ModelBlock model)
        {
            this.location = location;
            this.model = model;
        }

        public Collection<ResourceLocation> getDependencies()
        {
            if(model.func_178305_e() == null || model.func_178305_e().func_110623_a().startsWith("builtin/")) return Collections.emptyList();
            return Collections.singletonList(model.func_178305_e());
        }

        public Collection<ResourceLocation> getTextures()
        {
            // setting parent here to make textures resolve properly
            if(model.func_178305_e() != null)
            {
                if(model.func_178305_e().func_110623_a().equals("builtin/generated"))
                {
                    model.field_178315_d = field_177606_o;
                }
                else
                {
                    try
                    {
                        IModel parent = getModel(model.func_178305_e());
                        if(parent instanceof VanillaModelWrapper)
                        {
                            model.field_178315_d = ((VanillaModelWrapper) parent).model;
                        }
                        else
                        {
                            throw new IllegalStateException("vanilla model '" + model + "' can't have non-vanilla parent");
                        }
                    }
                    catch (IOException e)
                    {
                        FMLLog.warning("Could not load vanilla model parent '" + model.func_178305_e() + "' for '" + model + "': " + e.toString());
                        IModel missing = ModelLoader.this.getMissingModel();
                        if (missing instanceof VanillaModelWrapper)
                        {
                            model.field_178315_d = ((VanillaModelWrapper)missing).model;
                        }
                        else
                        {
                            throw new IllegalStateException("vanilla model '" + model + "' has missing parent, and missing model is not a vanilla model");
                        }
                    }
                }
            }

            ImmutableSet.Builder<ResourceLocation> builder = ImmutableSet.builder();

            if(func_177581_b(model))
            {
                for(String s : (List<String>)ItemModelGenerator.field_178398_a)
                {
                    String r = model.func_178308_c(s);
                    ResourceLocation loc = new ResourceLocation(r);
                    if(!r.equals(s))
                    {
                        builder.add(loc);
                    }
                    // mojang hardcode
                    if(model.func_178310_f() == field_177618_p && !loc.equals(TextureMap.field_174945_f))
                    {
                        TextureAtlasSprite.func_176603_b(loc.toString());
                    }
                    else if(model.func_178310_f() == field_177617_q && !loc.equals(TextureMap.field_174945_f))
                    {
                        TextureAtlasSprite.func_176602_a(loc.toString());
                    }
                }
            }
            for(String s : (Iterable<String>)model.field_178318_c.values())
            {
                if(!s.startsWith("#"))
                {
                    builder.add(new ResourceLocation(s));
                }
            }
            return builder.build();
        }

        public IFlexibleBakedModel bake(IModelState state, VertexFormat format, Function<ResourceLocation, TextureAtlasSprite> bakedTextureGetter)
        {
            if(!Attributes.moreSpecific(format, Attributes.DEFAULT_BAKED_FORMAT))
            {
                throw new IllegalArgumentException("can't bake vanilla models to the format that doesn't fit into the default one: " + format);
            }
            ModelBlock model = this.model;
            if(model == null) return getMissingModel().bake(state, format, bakedTextureGetter);
            ItemCameraTransforms transforms = model.func_181682_g();
            boolean uvlock = false;
            if(state instanceof UVLock)
            {
                uvlock = true;
                state = ((UVLock)state).getParent();
            }
            IPerspectiveState perState = state instanceof IPerspectiveState ? (IPerspectiveState)state : new IPerspectiveState.Impl(state, transforms);
            if(func_177581_b(model))
            {
                return new ItemLayerModel(model).bake(perState, format, bakedTextureGetter);
            }
            if(func_177587_c(model)) return new IFlexibleBakedModel.Wrapper(new BuiltInModel(transforms), format);
            // TODO perspective awareness for this
            return bakeNormal(model, perState, state.apply(this), format, bakedTextureGetter, uvlock);
        }

        private IFlexibleBakedModel bakeNormal(ModelBlock model, IPerspectiveState perState, TRSRTransformation state, VertexFormat format, Function<ResourceLocation, TextureAtlasSprite> bakedTextureGetter, boolean uvLocked)
        {
            TextureAtlasSprite particle = bakedTextureGetter.apply(new ResourceLocation(model.func_178308_c("particle")));
            SimpleBakedModel.Builder builder = (new SimpleBakedModel.Builder(model)).func_177646_a(particle);
            for(BlockPart part : (Iterable<BlockPart>)model.func_178298_a())
            {
                for(Map.Entry<EnumFacing, BlockPartFace> e : (Iterable<Map.Entry<EnumFacing, BlockPartFace>>)part.field_178240_c.entrySet())
                {
                    TextureAtlasSprite textureatlassprite1 = bakedTextureGetter.apply(new ResourceLocation(model.func_178308_c(e.getValue().field_178242_d)));

                    if (e.getValue().field_178244_b == null || !TRSRTransformation.isInteger(state.getMatrix()))
                    {
                        builder.func_177648_a(makeBakedQuad(part, e.getValue(), textureatlassprite1, e.getKey(), state, uvLocked));
                    }
                    else
                    {
                        builder.func_177650_a(state.rotate(e.getValue().field_178244_b), makeBakedQuad(part, e.getValue(), textureatlassprite1, e.getKey(), state, uvLocked));
                    }
                }
            }

            return new IPerspectiveAwareModel.MapWrapper(new IFlexibleBakedModel.Wrapper(builder.func_177645_b(), format), perState, this);
        }

        public IModelState getDefaultState()
        {
            return ModelRotation.X0_Y0;
        }

        @Override
        public IModel retexture(ImmutableMap<String, String> textures)
        {
            if (textures.isEmpty())
                return this;

            List<BlockPart> elements = Lists.newArrayList(); //We have to duplicate this so we can edit it below.
            for (BlockPart part : (List<BlockPart>)this.model.func_178298_a())
            {
                elements.add(new BlockPart(part.field_178241_a, part.field_178239_b, Maps.newHashMap(part.field_178240_c), part.field_178237_d, part.field_178238_e));
            }

            ModelBlock neweModel = new ModelBlock(this.model.func_178305_e(), elements,
                Maps.newHashMap(this.model.field_178318_c), this.model.func_178309_b(), this.model.func_178311_c(), //New Textures man VERY IMPORTANT
                model.func_181682_g());
            neweModel.field_178317_b = this.model.field_178317_b;
            neweModel.field_178315_d = this.model.field_178315_d;

            Set<String> removed = Sets.newHashSet();

            for (Entry<String, String> e : textures.entrySet())
            {
                if ("".equals(e.getValue()))
                {
                    removed.add(e.getKey());
                    neweModel.field_178318_c.remove(e.getKey());
                }
                else
                    neweModel.field_178318_c.put(e.getKey(), e.getValue());
            }

            // Map the model's texture references as if it was the parent of a model with the retexture map as its textures.
            Map<String, String> remapped = Maps.newHashMap();

            for (Entry<String, String> e : (Set<Entry<String, String>>)neweModel.field_178318_c.entrySet())
            {
                if (e.getValue().startsWith("#"))
                {
                    String key = e.getValue().substring(1);
                    if (neweModel.field_178318_c.containsKey(key))
                        remapped.put(e.getKey(), (String)neweModel.field_178318_c.get(key));
                }
            }

            neweModel.field_178318_c.putAll(remapped);

            //Remove any faces that use a null texture, this is for performance reasons, also allows some cool layering stuff.
            for (BlockPart part : (List<BlockPart>)neweModel.func_178298_a())
            {
                Iterator<Entry<EnumFacing, BlockPartFace>> itr = part.field_178240_c.entrySet().iterator();
                while (itr.hasNext())
                {
                    Entry<EnumFacing, BlockPartFace> entry = itr.next();
                    if (removed.contains(entry.getValue().field_178242_d))
                        itr.remove();
                }
            }

            return new VanillaModelWrapper(location, neweModel);
        }
    }

    public static class UVLock implements IModelState
    {
        private final IModelState parent;

        public UVLock(IModelState parent)
        {
            this.parent = parent;
        }

        public IModelState getParent()
        {
            return parent;
        }

        public TRSRTransformation apply(IModelPart part)
        {
            return parent.apply(part);
        }
    }

    // Weighted models can contain multiple copies of 1 model with different rotations - this is to make it work with IModelState (different copies will be different objects).
    private static class WeightedPartWrapper implements IModel
    {
        private final IModel model;

        public WeightedPartWrapper(IModel model)
        {
            this.model = model;
        }

        public Collection<ResourceLocation> getDependencies()
        {
            return model.getDependencies();
        }

        public Collection<ResourceLocation> getTextures()
        {
            return model.getTextures();
        }

        public IFlexibleBakedModel bake(IModelState state, VertexFormat format, Function<ResourceLocation, TextureAtlasSprite> bakedTextureGetter)
        {
            return model.bake(state, format, bakedTextureGetter);
        }

        public IModelState getDefaultState()
        {
            return model.getDefaultState();
        }
    }

    private class WeightedRandomModel implements IModel
    {
        private final List<Variant> variants;
        private final List<ResourceLocation> locations = new ArrayList<ResourceLocation>();
        private final List<IModel> models = new ArrayList<IModel>();
        private final IModelState defaultState;

        @Deprecated @SuppressWarnings("unused") public WeightedRandomModel(Variants variants){ this(null, variants); } // Remove 1.9
        public WeightedRandomModel(ModelResourceLocation parent, Variants variants)
        {
            this.variants = variants.func_178420_b();
            ImmutableMap.Builder<IModelPart, IModelState> builder = ImmutableMap.builder();
            for (Variant v : (List<Variant>)variants.func_178420_b())
            {
                ResourceLocation loc = v.func_178431_a();
                locations.add(loc);

                IModel model = null;
                try
                {
                    model = getModel(loc);
                }
                catch (Exception e)
                {
                    /*
                     * Vanilla eats this, which makes it only show variants that have models.
                     * But that doesn't help debugging, so we maintain the missing model
                     * so that resource pack makers have a hint that their states are broken.
                     */
                    FMLLog.warning("Unable to load block model: \'" + loc + "\' for variant: \'" + parent + "\': " + e.toString());
                    model = getMissingModel();
                }

                if (v instanceof ISmartVariant)
                {
                    model = ((ISmartVariant)v).process(model, ModelLoader.this);
                    textures.addAll(model.getTextures()); // Kick this, just in case.
                }

                model = new WeightedPartWrapper(model);
                models.add(model);
                builder.put(model, v.getState());
            }

            if (models.size() == 0) //If all variants are missing, add one with the missing model and default rotation.
            {
                IModel missing = getMissingModel();
                models.add(missing);
                builder.put(missing, TRSRTransformation.identity());
            }

            defaultState = new MapModelState(builder.build());

        }

        public Collection<ResourceLocation> getDependencies()
        {
            return ImmutableList.copyOf(locations);
        }

        public Collection<ResourceLocation> getTextures()
        {
            return Collections.emptyList();
        }

        private IModelState addUV(boolean uv, IModelState state)
        {
            if(uv) return new UVLock(state);
            return state;
        }

        private IModelState getState(IModelState state, IModelPart part)
        {
            if(state instanceof MapModelState)
            {
                return ((MapModelState)state).getState(part);
            }
            return state;
        }

        public IFlexibleBakedModel bake(IModelState state, VertexFormat format, Function<ResourceLocation, TextureAtlasSprite> bakedTextureGetter)
        {
            if(!Attributes.moreSpecific(format, Attributes.DEFAULT_BAKED_FORMAT))
            {
                throw new IllegalArgumentException("can't bake vanilla weighted models to the format that doesn't fit into the default one: " + format);
            }
            if(variants.size() == 1)
            {
                Variant v = variants.get(0);
                IModel model = models.get(0);
                return model.bake(addUV(v.func_178433_c(), getState(state, model)), format, bakedTextureGetter);
            }
            WeightedBakedModel.Builder builder = new WeightedBakedModel.Builder();
            for(int i = 0; i < variants.size(); i++)
            {
                IModel model = models.get(i);
                Variant v =  variants.get(i);
                builder.func_177677_a(model.bake(addUV(v.func_178433_c(), getState(state, model)), format, bakedTextureGetter), variants.get(i).func_178430_d());
            }
            return new FlexibleWeightedBakedModel(builder.func_177676_a(), Attributes.DEFAULT_BAKED_FORMAT);
        }

        public IModelState getDefaultState()
        {
            return defaultState;
        }
    }

    private static class FlexibleWeightedBakedModel extends WeightedBakedModel implements IFlexibleBakedModel
    {
        @SuppressWarnings("unused")
        private final WeightedBakedModel parent;
        private final VertexFormat format;

        public FlexibleWeightedBakedModel(WeightedBakedModel parent, VertexFormat format)
        {
            super(parent.field_177565_b);
            this.parent = parent;
            this.format = format;
        }

        public VertexFormat getFormat()
        {
            return format;
        }
    }

    @SuppressWarnings("unused")
    private boolean isBuiltinModel(ModelBlock model)
    {
        return model == field_177606_o || model == field_177618_p || model == field_177617_q || model == field_177616_r;
    }

    public IModel getMissingModel()
    {
        if (missingModel == null)
        {
            try
            {
                missingModel = getModel(new ResourceLocation(field_177604_a.func_110624_b(), field_177604_a.func_110623_a()));
            }
            catch (IOException e)
            {
                // If this ever happens things are bad. Should never NOT be able to load the missing model.
                Throwables.propagate(e);
            }
        }
        return missingModel;
    }

    public IModel getItemModel()
    {
        return itemModel;
    }

    static enum VanillaLoader implements ICustomModelLoader
    {
        instance;

        private ModelLoader loader;

        void setLoader(ModelLoader loader)
        {
            this.loader = loader;
        }

        ModelLoader getLoader()
        {
            return loader;
        }

        public void func_110549_a(IResourceManager resourceManager)
        {
            // do nothing, cause loader will store the reference to the resourceManager
        }

        public boolean accepts(ResourceLocation modelLocation)
        {
            return true;
        }

        public IModel loadModel(ResourceLocation modelLocation) throws IOException
        {
            return loader.new VanillaModelWrapper(modelLocation, loader.func_177594_c(modelLocation));
        }
    }

    public static class White extends TextureAtlasSprite
    {
        public static ResourceLocation loc = new ResourceLocation("white");
        public static White instance = new White();

        protected White()
        {
            super(loc.toString());
        }

        @Override
        public boolean hasCustomLoader(IResourceManager manager, ResourceLocation location)
        {
            return true;
        }

        @Override
        public boolean load(IResourceManager manager, ResourceLocation location)
        {
            BufferedImage image = new BufferedImage(16, 16, BufferedImage.TYPE_INT_ARGB);
            Graphics2D graphics = image.createGraphics();
            graphics.setBackground(Color.WHITE);
            graphics.clearRect(0, 0, 16, 16);
            BufferedImage[] images = new BufferedImage[Minecraft.func_71410_x().field_71474_y.field_151442_I + 1];
            images[0] = image;
            try
            {
                func_180598_a(images, null);
            }
            catch(IOException e)
            {
                throw new RuntimeException(e);
            }
            return false;
        }

        public void register(TextureMap map)
        {
            map.setTextureEntry(White.loc.toString(), White.instance);
        }
    }

    public void onPostBakeEvent(IRegistry<ModelResourceLocation, IBakedModel> modelRegistry)
    {
        IBakedModel missingModel = modelRegistry.func_82594_a(field_177604_a);
        for(ModelResourceLocation missing : missingVariants)
        {
            IBakedModel model = modelRegistry.func_82594_a(missing);
            if(model == null || model == missingModel)
            {
                FMLLog.severe("Model definition for location %s not found", missing);
            }
        }
        isLoading = false;
    }

    private static final Map<RegistryDelegate<Block>, IStateMapper> customStateMappers = Maps.newHashMap();

    public static void setCustomStateMapper(Block block, IStateMapper mapper)
    {
        customStateMappers.put(block.delegate, mapper);
    }

    public static void onRegisterAllBlocks(BlockModelShapes shapes)
    {
        for (Entry<RegistryDelegate<Block>, IStateMapper> e : customStateMappers.entrySet())
        {
            shapes.func_178121_a(e.getKey().get(), e.getValue());
        }
    }

    private static final Map<RegistryDelegate<Item>, ItemMeshDefinition> customMeshDefinitions = com.google.common.collect.Maps.newHashMap();
    private static final Map<Pair<RegistryDelegate<Item>, Integer>, ModelResourceLocation> customModels = com.google.common.collect.Maps.newHashMap();

    public static void setCustomModelResourceLocation(Item item, int metadata, ModelResourceLocation model)
    {
        customModels.put(Pair.of(item.delegate, metadata), model);
    }

    public static void setCustomMeshDefinition(Item item, ItemMeshDefinition meshDefinition)
    {
        customMeshDefinitions.put(item.delegate, meshDefinition);
    }

    public static void onRegisterItems(ItemModelMesher mesher)
    {
        for (Map.Entry<RegistryDelegate<Item>, ItemMeshDefinition> e : customMeshDefinitions.entrySet())
        {
            mesher.func_178080_a(e.getKey().get(), e.getValue());
        }
        for (Entry<Pair<RegistryDelegate<Item>, Integer>, ModelResourceLocation> e : customModels.entrySet())
        {
            mesher.func_178086_a(e.getKey().getLeft().get(), e.getKey().getRight(), e.getValue());
        }
    }
}
